/* PostgreSQL Python Interface 
 * ---------------------------
 * Copyright (c) 2000 - Eric BIANCHI & Thierry MICHEL 
 * Copyright (c) 2000 - MixadLive srl Torino, Italy 
 * http://www.mixadlive.com 
 *
 * This library is free software, you can redistribute it and/or 
 * modify it under the terms of the GNU Lesser General Public Licence as 
 * published by the Free Software Foundation. 
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU Lesser General Public License for more details 
 * You should have received a copy of the GNU Lesser General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. 
 *
 * MixadLive, Srl., hereby disclaims all copyright
 * interest in the library `PoPy' written  
 * by Eric BIANCHI and Thierry MICHEL 
 *
 * -- Eric - Developer 
 * -- Thierry - Developer
 * For bugs & comments: eric@mixadlive.com, thierry@mixadlive.com 
 *
 * This PostgreSQL python interface is Python DB-API version 2.0
 * compliant. For more informations about DB-API specifications 
 * you can find them at : 
 *
 * http://www.python.org 
 *
 */


#include <Python.h>
#include <structmember.h>
#include <postgresql/libpq-fe.h>
#include <postgresql/libpq/libpq-fs.h>
#include <regex.h>
#include <string.h>
#include <stdlib.h>

/* Handling Date and Time Objects */
#include <time.h>
#include "mx/mx.h"
#include "mx/mxDateTime.h"



static PyObject *PoPy_connect(PyObject *self,PyObject *args);
/* Must be cancelled */
static PyObject *PoPy_test(PyObject *string,PyObject *dict);
static PyObject *PoPy__BinaryObject_new (Oid oid,PGconn *conx);

#define PG_LO_NOTDEFINED -1000
#define PG_FALSE 0
#define PG_TRUE  1
#define MAX_BUFFER 1024
#define PG_LOCREAT 1
#define PG_LOUPDATE 2

/* Column status */
#define PG_NULL 24
#define PG_VARCHAR 259

/* PostgreSQL types definition */
#define PG_BOOLEAN 16
#define PG_FLOAT 700
#define PG_DOUBLE 701
#define PG_SHORT 21
#define PG_INTEGER 23
#define PG_DATE 1082
#define PG_TIME 1083
#define PG_TIMESTAMP 1084
#define PG_CASH 790
#define PG_BINARY 26
#define PG_ROWID 1000 /* Must be set bad value ! */


#define is_conx_object(ob) ((ob)->ob_type == &PoPy__ConnectionObject_Type)
#define Binary_Check(op) ((op)->ob_type == &PoPy__BinaryObject_Type)

static PyObject * PoPy_Error;

/********************************/
/* PG DB-API Objects definition */
/********************************/

/* Binary object */
typedef struct {
  PyObject_HEAD
  void *buffer;
  Oid oid;
  size_t size;
} PoPy__BinaryObject;

/* Cursor Object */
typedef struct {
  PyObject_HEAD
  int rowcount;
/*   int current; */
  int auto_commit;
  char cursor_name[15];
  int arraysize;
  char state;
  char executed;
  char last_query[MAX_BUFFER];
  char req_type;
  PyObject *datestyle;
  PyObject *description;
  PGconn *conn;
  PGresult *result;
} PoPy__CursorObject;

/* Connection Object */
typedef struct {
  PyObject_HEAD
  PGconn *conn;
  int open;
  int auto_commit;
  int commit_on;
/*   PyObject *cursor; */
  
} PoPy__ConnectionObject;

/* Exceptions definition */
static PyObject * PoPy_Warning;
static PyObject * PoPy_Error;
static PyObject * PoPy_InterfaceError;
static PyObject * PoPy_DatabaseError;
static PyObject * PoPy_DataError;
static PyObject * PoPy_OperationalError;
static PyObject * PoPy_IntegrityError;
static PyObject * PoPy_InternalError;
static PyObject * PoPy_ProgrammingError;
static PyObject * PoPy_NotSupportedError;


/* Types definition */
extern PyTypeObject PoPy__ConnectionObject_Type;
extern PyTypeObject PoPy__CursorObject_Type;
extern PyTypeObject PoPy__BinaryObject_Type;


/****************************/
/* Postgres types converter */
/****************************/
static PyObject *
PoPy_converter (result,tuple,field,conx)
     PGresult *result;
     int tuple;
     int field;
     PGconn *conx;
{
  int ftype;
  long int one=1, zero=0;

  ftype = PQftype(result,field);

  if(PQgetisnull(result,tuple,field)) {
    Py_INCREF(Py_None);
    return Py_None;
  }
    
  switch (ftype){
  case PG_BOOLEAN: return (!strcmp(PQgetvalue(result,tuple,field),"t")?PyInt_FromLong(one):PyInt_FromLong(zero));
  case PG_SHORT:
  case PG_ROWID:
  case PG_INTEGER:return PyInt_FromLong(atoi(PQgetvalue(result,tuple,field)));
  case PG_DATE:
  case PG_TIME:
  case PG_TIMESTAMP: return PyString_FromString(PQgetvalue(result,tuple,field));
  case PG_BINARY: return PoPy__BinaryObject_new(atoi(PQgetvalue(result,tuple,field)),conx) ;
  case PG_FLOAT:
  case PG_DOUBLE:return PyFloat_FromDouble(atof(PQgetvalue(result,tuple,field)));
  default: return PyString_FromString(PQgetvalue(result,tuple,field));
  }
}

int 
PoPy_LobDetector (conx,query,dict)
     PGconn *conx;
     char *query;
     PyObject *dict;
{
  int i;
  int _n_items,_flag=0,_res;
  regex_t *_regex;
  PyObject *_keys,*_key, *_value, *_values;
  PoPy__BinaryObject *_bo;

  char _pattern[MAX_BUFFER];
  char _tmpl[]="%(%s)s";
  int _fd;

  _keys = PyDict_Keys(dict);
  _values = PyDict_Values(dict);
  _n_items = PyList_Size(_keys);

  _regex = (regex_t *) malloc (sizeof(regex_t)); 
#ifdef __DEBUG__
  fprintf(stdout,"query: <%s>\n",query); 
#endif
  for(i=0; i<_n_items; i++)
    {
      _key = PyList_GetItem(_keys,i);
      _value = PyList_GetItem(_values,i);
      if (Binary_Check(_value)) {
	_bo = (PoPy__BinaryObject *) _value;
	sprintf(_pattern,_tmpl,PyString_AS_STRING(_key));
	regcomp(_regex,_pattern,REG_ICASE | REG_NOSUB); 
	if(!(_res = regexec(_regex,query,0,NULL,REG_NOTBOL))) {
	  if (_bo->oid == -1) {
	    _bo->oid = lo_creat(conx,INV_READ | INV_WRITE);
	    _fd = lo_open(conx,_bo->oid,INV_READ | INV_WRITE);
	    lo_write(conx,_fd,_bo->buffer,_bo->size);
	    lo_close(conx,_fd);
	  }
	  _flag++;
	}
      }
    }
  if (_flag)
    regfree(_regex);

  return _flag;
}

/* Return 1 if the query is an select query */
int 
PoPy_selectQuery (query)
     const char *query;
{
  int i;
  for(i = 0;i < strlen(query);i++)
    {
      if(query[i] != ' ')
        {
          if(strncmp(query+i,"select",6) == 0) return 1;
          else return 0;
        }
    }
}

/* ************************************************ */
/*         BINARY OBJECT METHODS DECLARATION        */
/* ************************************************ */

PyObject *
PoPy__BinaryObject_repr (self)
     PoPy__BinaryObject *self;
{
  char buffer[MAX_BUFFER];

  sprintf(buffer,"%d",self->oid);
  return Py_BuildValue ("s",buffer);
}

void
PoPy__BinaryObject_dealloc (self)
     PoPy__BinaryObject *self;
{
  free(self->buffer);
  PyMem_Free((char *)self);
}

PyObject *
PoPy__BinaryObject_size (self)
     PoPy__BinaryObject *self;
{
  return PyInt_FromLong(self->size);
}

static PyObject *
PoPy__BinaryObject_new (oid,conx)
     PGconn *conx;
     Oid oid;
{
  PoPy__BinaryObject *_bobj;
  size_t size;
  int _fd;
  
  if(!(_bobj = PyObject_NEW(PoPy__BinaryObject,&PoPy__BinaryObject_Type))) return NULL;

  if((_fd = lo_open(conx,oid,INV_READ)) == -1)
    {
      Py_INCREF(Py_None);
      return Py_None;
    }

  lo_lseek(conx,_fd,SEEK_CUR,SEEK_END);
  _bobj->size = lo_tell(conx,_fd) - 1;
  lo_lseek(conx,_fd,0,SEEK_SET);
  _bobj->buffer = (void *) malloc (_bobj->size);

  if((size = lo_read(conx,_fd,_bobj->buffer,_bobj->size)) < 0)
    {
      Py_INCREF(Py_None);
      return Py_None;
    }

  lo_close(conx,_fd);
  _bobj->ob_type = &PoPy__BinaryObject_Type;
  _bobj->oid = oid;

  return (PyObject *)_bobj;
}

static struct memberlist PoPy__BinaryObject_memberlist[] = {
	{"size", T_INT, 0, RO},
	{"buffer", T_OBJECT, 0, RO},
	{NULL} /* Sentinel */
};

static PyObject *
PoPy__BinaryObject_getattr(self, name)
     PoPy__BinaryObject *self;
     char *name;
{
  if(!strcmp(name,"size"))
    return PyInt_FromLong(self->size);
  if(!strcmp(name,"buffer"))
    return PyString_FromStringAndSize(self->buffer, self->size);
  return PyMember_Get((char *)self, PoPy__BinaryObject_memberlist, name);
}

static PyObject *
PoPy__BinaryObject_NEW (self,args)
     PyObject *self;
     PyObject *args;
{
  PoPy__BinaryObject *_bobj;
  char *_data=NULL;
  int size;

  if(!(_bobj = PyObject_NEW(PoPy__BinaryObject,&PoPy__BinaryObject_Type))) return NULL;

  if(!PyArg_ParseTuple(args,"s#",&_data,&size))
    return NULL;

  _bobj->buffer = (void *) malloc (size);
  if(!memcpy(_bobj->buffer,_data,size))
    {
      Py_INCREF(Py_None);
      return Py_None;
    }

  _bobj->ob_type = &PoPy__BinaryObject_Type;
  _bobj->oid = -1;
  _bobj->size = size;
  return (PyObject *)_bobj;
}

/* ************************************************ */
/*         CURSOR OBJECT METHODS DECLARATION        */
/* ************************************************ */

static char PoPy__CursorObject__doc__ []="PoPy";

static PyObject *
PoPy__CursorObject_close(self)
     PoPy__CursorObject *self;
{
  PGresult *_result=NULL;
  char buffer[MAX_BUFFER];

  if(self->executed) {
    sprintf(buffer,"CLOSE %s",self->cursor_name);
    _result = PQexec(self->conn,buffer);
    if((PQresultStatus(_result)) == PGRES_FATAL_ERROR)
      {
	PyErr_SetString(PoPy_OperationalError,PQresultErrorMessage(_result));
	PyErr_ExceptionMatches(PoPy_OperationalError);
	return NULL;
      }
    PQclear(_result);
    self->executed = 0;
  }
  self->state = 0;
  Py_INCREF(Py_None);
  return Py_None;
}
 
static PyObject *
PoPy__CursorObject_description(self,nfields)
     PoPy__CursorObject *self;
{
  int i;
  int _fsize;
  PyObject *fname,*ftype,*fsize,*fnull,*r;  

  Py_XDECREF(self->description);
  if(PQresultStatus(self->result) == PGRES_TUPLES_OK)
    {
      if (!(self->description = PyList_New(0))) return NULL;
      
      if((fname=(PyObject*)PyTuple_New(nfields))==NULL) return NULL;
      if((ftype=(PyObject*)PyTuple_New(nfields))==NULL) return NULL;      
      if((fsize=(PyObject*)PyTuple_New(nfields))==NULL) return NULL;      
      if((fnull=(PyObject*)PyTuple_New(nfields))==NULL) return NULL;

      for(i=0; i<nfields; i++)
	{
	  PyObject *value,*value1;
	  value = PyString_FromString(PQfname(self->result,i));
	  PyTuple_SET_ITEM (fname,i,value);
	  switch(PQftype(self->result,i)) {
	  case PG_BOOLEAN:
	  case PG_SHORT:
	  case PG_INTEGER: 
	  case PG_FLOAT:
	  case PG_DOUBLE:value1=PyString_FromString("NUMBER");break;
	  case PG_ROWID:value1=PyString_FromString("ROWID");break;
	  case PG_DATE:
	  case PG_TIME: value1=PyString_FromString("DATETIME");break;
	  case PG_BINARY: value1=PyString_FromString("BINARY");break;
	  default: value1=PyString_FromString("STRING");break;
	  }
      
	  PyTuple_SET_ITEM (ftype,i,value1);
	  _fsize = PQfsize(self->result,i);
	  if (_fsize != -1)
	    PyTuple_SET_ITEM (fsize,i,PyInt_FromLong(PQfsize(self->result,i)));
	  else
	    PyTuple_SET_ITEM (fsize,i,PyInt_FromLong(PQgetlength(self->result,0,i)));
	  PyTuple_SET_ITEM(fnull,i,PQfmod(self->result,i)==PG_NULL?PyString_FromString("NULL"):PyString_FromString("NOT NULL"));
	  
	}
      PyList_Append(self->description, fname);
      Py_DECREF(fname);
      PyList_Append(self->description, ftype);
      Py_DECREF(ftype);
      PyList_Append(self->description, fsize);
      Py_DECREF(fsize);
      PyList_Append(self->description, fnull);
      Py_DECREF(fnull);
    }
  Py_INCREF(Py_None);
  return Py_None;
}

  
static PyObject *
PoPy__CursorObject_fetchone(self)
     PoPy__CursorObject *self;
{
  int n_fields,n_tuples,i;
  PyObject *row,*r;
  char buffer[MAX_BUFFER];

  if(!self->state) {
    PyErr_SetString(PoPy_InternalError,"Cursor closed");
    PyErr_ExceptionMatches(PoPy_InternalError);
    Py_INCREF(Py_None);
    return Py_None;
  }

  if(self->executed == PG_FALSE) {
    PyErr_SetString(PoPy_OperationalError,"select query not executed");
    PyErr_ExceptionMatches(PoPy_OperationalError);
    Py_INCREF(Py_None);
    return Py_None;
  }

  if(self->auto_commit == PG_FALSE)
    {
      sprintf(buffer,"FETCH FORWARD 1 IN %s",self->cursor_name);
      if(self->result != NULL)
	PQclear(self->result);
      self->result = PQexec(self->conn,buffer);
      if((PQresultStatus(self->result)) == PGRES_FATAL_ERROR)
	{
	  PyErr_SetString(PoPy_OperationalError,PQresultErrorMessage(self->result));
	  PyErr_ExceptionMatches(PoPy_OperationalError);
	  return NULL;
	}

    }
  else {
    PyErr_SetString(PoPy_ProgrammingError,"auto-commit mode is ON");
    PyErr_ExceptionMatches(PoPy_ProgrammingError);
    Py_INCREF(Py_None);
    return Py_None;
  }

  n_fields = PQnfields(self->result);
  n_tuples = PQntuples(self->result);

  if(n_tuples == 0) {
    PyErr_SetString(PoPy_DatabaseError,"No more results to fetch\n");
    PyErr_ExceptionMatches(PoPy_DatabaseError);
    return NULL;
  }

  if(self->rowcount == -1)
    self->rowcount = 0;
  self->rowcount += n_tuples;

  if(self->description == Py_None) 
     PoPy__CursorObject_description(self,n_fields);

  if (!(r = PyList_New(0))) return NULL;
  if((row=(PyObject*)PyTuple_New(n_fields))==NULL) return NULL;

  for(i=0; i<n_fields; i++)
    {
      PyObject *value;
      value = PoPy_converter(self->result,0,i,self->conn);
/*       value = PyString_FromString(PQgetvalue(res,0,i)); */
      PyTuple_SET_ITEM (row,i,value);
    }
  PyList_Append(r, row);
  Py_DECREF(row);

  return r;
}
  
static PyObject *
PoPy__CursorObject_fetchall (self)
     PoPy__CursorObject *self;
{
  PyObject *r,*v;
  int i,t,n_fields,n_tuples;
  char buffer[MAX_BUFFER];

  if(!self->state) {
    PyErr_SetString(PoPy_InternalError,"Cursor closed");
    PyErr_ExceptionMatches(PoPy_InternalError);
    Py_INCREF(Py_None);
    return Py_None;
  }

/*   if(self->executed == PG_FALSE) { */
/*     PyErr_SetString(PoPy_OperationalError,"select query not executed"); */
/*     PyErr_ExceptionMatches(PoPy_OperationalError); */
/*     Py_INCREF(Py_None); */
/*     return Py_None; */
/*   } */


  if(self->auto_commit == PG_FALSE)
    {
      sprintf(buffer,"FETCH FORWARD ALL IN %s",self->cursor_name);
      if(self->result != NULL)
	PQclear(self->result);
      self->result = PQexec(self->conn,buffer);
      if((PQresultStatus(self->result)) == PGRES_FATAL_ERROR)
	{
	  PyErr_SetString(PoPy_OperationalError,PQresultErrorMessage(self->result));
	  PyErr_ExceptionMatches(PoPy_OperationalError);
	  return NULL;
	}
    }
  else if ((self->executed == PG_FALSE) && !self->result) 
    {
      PyErr_SetString(PoPy_ProgrammingError,"fetch with no result");
      PyErr_ExceptionMatches(PoPy_ProgrammingError);
      Py_INCREF(Py_None);
      return Py_None;
    }
  
  n_fields = PQnfields(self->result);
  n_tuples = PQntuples(self->result);

  if(n_tuples == 0) {
    PyErr_SetString(PoPy_DatabaseError,"No more results to fetch\n");
    PyErr_ExceptionMatches(PoPy_DatabaseError);
    return NULL;
  }
  
  if(self->rowcount == -1)
    self->rowcount = 0;
  self->rowcount += n_tuples;

  if(self->description == Py_None) 
    PoPy__CursorObject_description(self,n_fields);

  if (!(r = PyList_New(0))) return NULL;

  for(t=0;t<n_tuples;t++) {
    if((v=(PyObject*)PyTuple_New(n_fields))==NULL) return NULL;
    
    for(i=0; i<n_fields; i++)
      {
	PyObject *value;
	value = PoPy_converter(self->result,t,i,self->conn);
	PyTuple_SET_ITEM (v,i,value);
      }
    PyList_Append(r, v);
    Py_DECREF(v);
  }
  return r;
}

static PyObject *
PoPy__CursorObject_fetchmany (self,args)
     PoPy__CursorObject *self;
     PyObject *args;
{
  PyObject *r,*v;
  int i,t,n_fields;
  unsigned int n_tuples;
  char buffer[MAX_BUFFER];

  if(!self->state) {
    PyErr_SetString(PoPy_InternalError,"Cursor closed");
    PyErr_ExceptionMatches(PoPy_InternalError);
    Py_INCREF(Py_None);
    return Py_None;
  }

  if(self->executed == PG_FALSE) {
    PyErr_SetString(PoPy_OperationalError,"select query not executed");
    PyErr_ExceptionMatches(PoPy_OperationalError);
    Py_INCREF(Py_None);
    return Py_None;
  }

  if (PyArg_NoArgs(args)) n_tuples = self->arraysize;
  else
    if(!PyArg_ParseTuple(args,"i",&n_tuples)) return NULL;

  if(self->auto_commit == PG_FALSE)
    {
      sprintf(buffer,"FETCH FORWARD %d IN %s",n_tuples,self->cursor_name);
      self->result = PQexec(self->conn,buffer);
      if((PQresultStatus(self->result)) == PGRES_FATAL_ERROR)
	{
	  PyErr_SetString(PoPy_OperationalError,PQresultErrorMessage(self->result));
	  PyErr_ExceptionMatches(PoPy_OperationalError);
	  return NULL;
	}
    }
  else {
    PyErr_SetString(PoPy_ProgrammingError,"auto-commit mode is ON");
    PyErr_ExceptionMatches(PoPy_ProgrammingError);
    Py_INCREF(Py_None);
    return Py_None;
  }

  n_fields = PQnfields(self->result);
  n_tuples = PQntuples(self->result);

  if(n_tuples == 0) {
    PyErr_SetString(PoPy_DatabaseError,"No more results to fetch\n");
    PyErr_ExceptionMatches(PoPy_DatabaseError);
    return NULL;
  }
  
  if(self->rowcount == -1)
    self->rowcount = 0;
  self->rowcount += n_tuples;
  
  if(self->description == Py_None) 
    PoPy__CursorObject_description(self,n_fields);

  if (!(r = PyList_New(0))) return NULL;

  for(t=0;t<n_tuples;t++) {
    if((v=(PyObject*)PyTuple_New(n_fields))==NULL) return NULL;
    
    for(i=0; i<n_fields; i++)
      {
	PyObject *value;
	value = PoPy_converter(self->result,t,i,self->conn);
/* 	value = PyString_FromString(PQgetvalue(self->result,t,i)); */
	PyTuple_SET_ITEM (v,i,value);
      }
    PyList_Append(r, v);
    Py_DECREF(v);
  }
  return r;
}

static PyObject *
PoPy__CursorObject_callproc (self,args)
     PoPy__CursorObject *self;
     PyObject *args;
{
  int _f;
  char *query;
  char buffer[MAX_BUFFER];
  PGresult *_result;
  char query_type;
  PyObject *_string=NULL,*_dict=NULL,*_query=NULL;
  PyObject *_var;


  if(!self->state) {
    PyErr_SetString(PoPy_InternalError,"Cursor closed");
    PyErr_ExceptionMatches(PoPy_InternalError);
    Py_INCREF(Py_None);
    return Py_None;
  }

/*   Get objects in the argument  */
  if(!PyArg_ParseTuple(args,"S|O!",&_string,&PyTuple_Type,&_dict))
    return NULL;

  /*   Set globals variables to eval the expression */
  if (_dict) {
    _var = PyDict_New();

    Py_INCREF(_var);
    PyDict_SetItemString(_var,"__builtins__",PyEval_GetBuiltins());
    PyDict_SetItemString(_var,"procedure",_string);
    PyDict_SetItemString(_var,"parameters",_dict);

    /*   execute the expression and get the result */
    
    if (PyTuple_Size(_dict) > 1) {
      PyRun_String("procedure = 'select %s%s' % (procedure,parameters)",Py_single_input,_var,_var);
    }
    else 
      PyRun_String("procedure = \"select %s('%s')\" % (procedure,parameters[0])",Py_single_input,_var,_var);
    _query = PyDict_GetItemString(_var,"procedure");
    Py_DECREF(_var);
  }
  else {
    _var = PyDict_New();
    PyDict_SetItemString(_var,"__builtins__",PyEval_GetBuiltins());
    PyDict_SetItemString(_var,"procedure",_string);
    Py_INCREF(_var);
    
    PyRun_String("procedure = 'select %s()' % (procedure)",Py_single_input,_var,_var);

    _query = PyDict_GetItemString(_var,"procedure");

    Py_DECREF(_var);
  }

  /* Cancel last results */
  if(self->result != NULL) {
    PQclear(self->result);
    self->result = NULL;
  }
  sprintf(buffer,"DECLARE %s CURSOR FOR %s",self->cursor_name,PyString_AS_STRING(_query));
  _result = PQexec(self->conn,buffer);

  if((PQresultStatus(_result)) == PGRES_FATAL_ERROR)
    {
      PyErr_SetString(PoPy_OperationalError,PQresultErrorMessage(_result));
      PyErr_ExceptionMatches(PoPy_OperationalError);
      return NULL;
    }
  PQclear(_result);

  self->executed = PG_TRUE;
  Py_INCREF(Py_None);
  return Py_None;
}


static PyObject *
PoPy__CursorObject_execute (self, args)
     PoPy__CursorObject *self;
     PyObject *args;
{
  int _f;
  char *query;
  char buffer[MAX_BUFFER];
  PGresult *_result;
  char query_type;
  PyObject *_string=NULL,*_dict=NULL,*_query=NULL;
  PyObject *_var;

  if(!self->state) {
    PyErr_SetString(PoPy_InternalError,"Cursor closed");
    PyErr_ExceptionMatches(PoPy_InternalError);
    Py_INCREF(Py_None);
    return Py_None;
  }

/*   Get objects in the argument  */
  if(!PyArg_ParseTuple(args,"S|O!",&_string,&PyDict_Type,&_dict))
    return NULL;
 


  self->req_type = PoPy_selectQuery(PyString_AS_STRING(_string)); 
  
  if(_dict) {
    /* Some tests */
#ifdef __DEBUG__
    printf("Calling LobDetector v0.3\n");
#endif
    if(!self->req_type) {
      PoPy_LobDetector (self->conn,PyString_AS_STRING(_string),_dict);
    }
    
    /*   Set globals variables to eval the expression */
    _var = PyDict_New();
    PyDict_SetItemString(_var,"__builtins__",PyEval_GetBuiltins());
    PyDict_SetItemString(_var,"format",_string);
    PyDict_SetItemString(_var,"dict",_dict);
    Py_INCREF(_var);

    /*   execute the expression and get the result */

    PyRun_String("format = format % dict",Py_single_input,_var,_var);

    _query = PyDict_GetItemString(_var,"format");
#ifdef __DEBUG__
    printf("REQUEST: <%s>\n",PyString_AS_STRING(_query));
#endif
    Py_DECREF(_var);
  }

  /* Cancel last results */
  if(self->result != NULL) {
    PQclear(self->result);
    self->result = NULL;
  }
  
  if(self->description != Py_None)
    {
      Py_XDECREF(self->description);
      Py_INCREF(Py_None);
      self->description = Py_None;
    }

  if((self->auto_commit == PG_FALSE) && (self->req_type))
    {
      if(self->executed == PG_TRUE)
	{
	  PyErr_SetString(PoPy_OperationalError,"Cursor already opened");
	  PyErr_ExceptionMatches(PoPy_OperationalError);
	  return NULL;
	}

      sprintf(buffer,"DECLARE %s CURSOR FOR %s",self->cursor_name,_query?PyString_AS_STRING(_query):PyString_AS_STRING(_string));
      _result = PQexec(self->conn,buffer);

      if((PQresultStatus(_result)) == PGRES_FATAL_ERROR)
	{
	  PyErr_SetString(PoPy_OperationalError,PQresultErrorMessage(_result));
	  PyErr_ExceptionMatches(PoPy_OperationalError);
	  return NULL;
	}
      PQclear(_result);
#ifdef __DEBUG__
      printf("cursor created !\n");
#endif
/*        PoPy__CursorObject_description(self); */
/*       self->rowcount = PQntuples(self->result); */

      self->executed = PG_TRUE;
    }
  else
    {
      self->result = PQexec(self->conn,_query?PyString_AS_STRING(_query):PyString_AS_STRING(_string));
      if((PQresultStatus(self->result)) == PGRES_FATAL_ERROR)
	{
	  PyErr_SetString(PoPy_OperationalError,PQresultErrorMessage(self->result));
	  PyErr_ExceptionMatches(PoPy_OperationalError);
	  return NULL;
	}
      if((PQresultStatus(self->result)) == PGRES_COMMAND_OK)
	{
	  PQclear(self->result);
	  self->result = NULL;
	}
      self->executed = PG_FALSE;
/*       self->current = 0; */
    }
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject *
PoPy__CursorObject_executemany(self,args)
     PoPy__CursorObject *self;
     PyObject *args;
{
  PyObject *_list,*_item,*_query,*_var,*_ret;
  int i,n;

  printf("executemany \n");
  if(!(PyArg_ParseTuple(args,"SO",&_query,&_list))) {
    PyErr_SetString(PoPy_ProgrammingError,"Bad argument type");
    PyErr_ExceptionMatches(PoPy_ProgrammingError);
    Py_INCREF(Py_None);
    return Py_None;
  }

  if(!PySequence_Check(_list)) {
    PyErr_SetString(PoPy_ProgrammingError,"Bad argument type");
    PyErr_ExceptionMatches(PoPy_ProgrammingError);
    Py_INCREF(Py_None);
    return Py_None;
      }
  
  n = PySequence_Length(_list);

  printf("number of items: %d\n",n);

  for(i=0; i<n; i++) {

    _item = PyList_GetItem(_list,i);

    if (!PyMapping_Check(_item)) {
      PyErr_SetString(PoPy_ProgrammingError,"Item is not a mapping");
      PyErr_ExceptionMatches(PoPy_ProgrammingError);
      Py_INCREF(Py_None);
      return Py_None;
    }

    _ret = PyObject_CallMethod((PyObject *)self,"execute","SO",_query,_item);
    if(_ret == NULL)
      {
 	PyErr_SetString(PoPy_OperationalError,"Excecute many error");
 	PyErr_ExceptionMatches(PoPy_OperationalError);
 	Py_INCREF(Py_None);
 	return Py_None;
      }
  }

  Py_INCREF(Py_None);
  return Py_None;
}
  

static PoPy__CursorObject*
PoPy__CursorObject_NEW (conn,auto_commit)
     PGconn *conn;
     int auto_commit;
{
  PoPy__CursorObject *c;
  PGresult *res;

  if (!(c = PyObject_NEW(PoPy__CursorObject, &PoPy__CursorObject_Type)))
    {
      PyErr_SetString(PoPy_InterfaceError,"failed to create a cursor");
      PyErr_ExceptionMatches(PoPy_InterfaceError);
      return NULL;
    }

  c->conn = conn;
  c->rowcount = -1;
  /*   c->current = -1; */
  sprintf(c->cursor_name,"Q%lx",(long)c);
  c->auto_commit = auto_commit;
  Py_INCREF(Py_None);
  c->description = Py_None;
  c->result = NULL;
  c->last_query[0] = '\0'; /* no query */
  c->state = 1;   
  c->arraysize = 1;
  c->executed = 0;
  c->ob_type = &PoPy__CursorObject_Type;
  return c;
}

/* Freeing the _ConnectionObject and close the connexion */
/* if it's opened                                       */
static void
PoPy__CursorObject_dealloc(self)
	PoPy__CursorObject *self;
{
  PQclear(self->result);
  Py_XDECREF(self->description);
/*   PyMem_Free((char *) self->description); */
  PyMem_Free((char *) self);

}

  
/* String representation of the _ConnectionObject */
PyObject *
PoPy__CursorObject_repr(self)
	PoPy__CursorObject *self;
{
  char buf[300];
	
  sprintf(buf, "<_CursorObject at %lx>\n",
	  (long)self);
  return PyString_FromString(buf);
}

static PyMethodDef PoPy__CursorObject_methods[] = {
  {"callproc",(PyCFunction)PoPy__CursorObject_callproc,METH_VARARGS},
  {"execute",(PyCFunction)PoPy__CursorObject_execute,METH_VARARGS},
  {"executemany",(PyCFunction)PoPy__CursorObject_executemany,METH_VARARGS},
  {"fetchone",(PyCFunction)PoPy__CursorObject_fetchone,0},
  {"fetchall",(PyCFunction)PoPy__CursorObject_fetchall,0},
  {"fetchmany",(PyCFunction)PoPy__CursorObject_fetchmany,METH_VARARGS},
  {"close",(PyCFunction)PoPy__CursorObject_close,0},
  {NULL, NULL}
};
      
static struct memberlist PoPy__CursorObject_memberlist[] = {
	{"description", T_OBJECT, offsetof(PoPy__CursorObject,description)},
	{"arraysize", T_INT, 0, 0},
	{"rowcount", T_INT, 0, RO},
	{"datestyle",T_OBJECT,0,RO},
	{NULL} /* Sentinel */
};

static int
PoPy__CursorObject_setattr(self, name, v)
	PoPy__CursorObject *self;
	char *name;
	PyObject *v;
{
  if (v == NULL) {
    PyErr_SetString(PyExc_AttributeError,
		    "can't delete cursor attributes");
    return -1;
  }
  if (!strcmp(name, "arraysize")) {
    if (!PyInt_Check(v)) {
      PyErr_SetString(PyExc_TypeError, "arraysize must be integer.");
      return -1;
    }
    self->arraysize = PyInt_AsLong(v);
    return 0;
  }

  /* unknown attribute */
  PyErr_SetString(PyExc_TypeError, "not a writable attribute.");
  return -1;
}


static PyObject *
PoPy__CursorObject_getattr(self, name)
     PoPy__CursorObject *self;
     char *name;
{
  PyObject *res;

  res = Py_FindMethod(PoPy__CursorObject_methods, (PyObject *)self, name);
  if (res != NULL)
    return res;
  PyErr_Clear();
  if (strcmp(name, "rowcount") == 0)
    return PyInt_FromLong((long)(self->rowcount));
  if (strcmp(name, "arraysize") == 0)
    return PyInt_FromLong((long)(self->arraysize));
  if (strcmp(name, "datestyle") == 0)
    return self->datestyle;
  return PyMember_Get((char *)self, PoPy__CursorObject_memberlist, name);
}

/* ************************************************ */
/*      CONNECTION  OBJECT METHODS DECLARATION      */
/* ************************************************ */

PoPy__ConnectionObject *
PoPy__ConnectionObject_NEW (args)
     PyObject *args;
{
  PGconn *conn;
  char *_arg0;
  PGresult *res;
  PoPy__ConnectionObject *c;

  if(!(c = PyObject_NEW(PoPy__ConnectionObject, &PoPy__ConnectionObject_Type))) return NULL;

  if (!PyArg_ParseTuple(args, "s",&_arg0))
      return NULL;

  c->conn = PQconnectdb(_arg0);

  if(PQstatus(c->conn) == CONNECTION_BAD)
    {
      PyErr_SetString(PoPy_DatabaseError,"Connection failed");
      PyErr_ExceptionMatches(PoPy_DatabaseError);
      return NULL;
    }

  c->auto_commit = PG_FALSE;
  c->commit_on = PG_FALSE;
  c->open = 1;
  c->ob_type = &PoPy__ConnectionObject_Type;

  res = PQexec(c->conn,"BEGIN");
  if((PQresultStatus(res)) == PGRES_FATAL_ERROR)
    {
      PyErr_SetString(PoPy_OperationalError,PQresultErrorMessage(res));
      PyErr_ExceptionMatches(PoPy_OperationalError);
      return NULL;
    }
  PQclear(res);
  res = PQexec(c->conn,"SET TRANSACTION ISOLATION LEVEL SERIALIZABLE");
  if((PQresultStatus(res)) == PGRES_FATAL_ERROR)
    {
      PyErr_SetString(PoPy_OperationalError,PQresultErrorMessage(res));
      PyErr_ExceptionMatches(PoPy_OperationalError);
      return NULL;
    }
  PQclear(res);
  return c;
}

/* String representation of the _ConnectionObject */
PyObject *
PoPy__ConnectionObject_repr(self)
	PoPy__ConnectionObject *self;
{
  char buf[300];
	
  if(self->open && PQstatus(self->conn)==CONNECTION_OK)
    {
      sprintf(buf, "%s connection to '%.256s' at %lx\nauto-commit: %s",
	      self->open ? "open" : "closed",
	      PQhost(self->conn),
	      (long)self,
	      self->auto_commit ? "ON" : "OFF");
      return PyString_FromString(buf);
    }
  sprintf(buf,"<Instance at %lx state: closed>",(long)self);
  return PyString_FromString(buf);
}

/* Freeing the _ConnectionObject and close the connexion */
/* if it's opened                                       */
static void
PoPy__ConnectionObject_dealloc(self)
	PoPy__ConnectionObject *self;
{
	if (self->open) {
	  PQfinish(self->conn);
	}
	PyMem_Free((char *) self);
}


static char PoPy_autoCommit__doc__ []="Allow to change the auto-commit mode (default: false)";

static PyObject*
PoPy__ConnectionObject_autoCommit(self,args)
     PoPy__ConnectionObject *self;
     PyObject *args;
{
  PGresult *res;
  int _value = 1;

  if (!PyArg_ParseTuple(args, "|i",&_value))
      return NULL;

  _value = (_value > 0?PG_TRUE:PG_FALSE);

  if (self->auto_commit != _value)
    {
      if (self->auto_commit==PG_FALSE)
	{
	  res = PQexec(self->conn,"END");
	  if((PQresultStatus(res)) == PGRES_FATAL_ERROR)
	    {
	      PyErr_SetString(PoPy_OperationalError,PQresultErrorMessage(res));
	      PyErr_ExceptionMatches(PoPy_OperationalError);
	      return NULL;
	    }
	  PQclear(res);
	}
      else {
	res = PQexec(self->conn,"BEGIN");
	if((PQresultStatus(res)) == PGRES_FATAL_ERROR)
	  {
	    PyErr_SetString(PoPy_OperationalError,PQresultErrorMessage(res));
	    PyErr_ExceptionMatches(PoPy_OperationalError);
	    return NULL;
	  }
	PQclear(res);
      }
    }
  
  self->auto_commit = _value;

  Py_INCREF(Py_None);
  return Py_None;
}
  
PyObject*
PoPy__ConnectionObject_rollback(self)
     PoPy__ConnectionObject *self;
{
  PGresult *_result=NULL;

  if (self->auto_commit==PG_FALSE && self->open==PG_TRUE && self->commit_on==PG_FALSE)
    {
      _result = PQexec(self->conn,"rollback");
      if((PQresultStatus(_result)) == PGRES_FATAL_ERROR)
	{
	  PyErr_SetString(PoPy_OperationalError,PQresultErrorMessage(_result));
	  PyErr_ExceptionMatches(PoPy_OperationalError);
	  return NULL;
	}
      PQclear(_result);
      _result = PQexec(self->conn,"BEGIN");
      if((PQresultStatus(_result)) == PGRES_FATAL_ERROR)
	{
	  PyErr_SetString(PoPy_OperationalError,PQresultErrorMessage(_result));
	  PyErr_ExceptionMatches(PoPy_OperationalError);
	  return NULL;
	}
      PQclear(_result);
      _result = PQexec(self->conn,"SET TRANSACTION ISOLATION LEVEL SERIALIZABLE");
      if((PQresultStatus(_result)) == PGRES_FATAL_ERROR)
	{
	  PyErr_SetString(PoPy_OperationalError,PQresultErrorMessage(_result));
	  PyErr_ExceptionMatches(PoPy_OperationalError);
	  return NULL;
	}
      self->commit_on = PG_FALSE;
      PQclear(_result);
    }
  else {
    PyErr_SetString(PoPy_NotSupportedError,"auto-commit mode is ON");
    PyErr_ExceptionMatches(PoPy_Error);
    return NULL;
  }
  Py_INCREF(Py_None);
  return Py_None;
}


static char PoPy_commit__doc__ []="The commit function";

PyObject *
PoPy__ConnectionObject_commit(self)
     PoPy__ConnectionObject *self;
{
  PGresult *_result=NULL;


  if(self->auto_commit==PG_FALSE && self->open==PG_TRUE)
    {
      _result =PQexec(self->conn,"commit");
      if((PQresultStatus(_result)) == PGRES_FATAL_ERROR)
	{
	  PyErr_SetString(PoPy_OperationalError,PQresultErrorMessage(_result));
	  PyErr_ExceptionMatches(PoPy_Error);
	  return NULL;
	}
      PQclear(_result);
      _result = PQexec(self->conn,"BEGIN");
      if((PQresultStatus(_result)) == PGRES_FATAL_ERROR)
	{
	  PyErr_SetString(PoPy_OperationalError,PQresultErrorMessage(_result));
	  PyErr_ExceptionMatches(PoPy_OperationalError);
	  return NULL;
	}
      PQclear(_result);
      _result = PQexec(self->conn,"SET TRANSACTION ISOLATION LEVEL SERIALIZABLE");
      if((PQresultStatus(_result)) == PGRES_FATAL_ERROR)
	{
	  PyErr_SetString(PoPy_OperationalError,PQresultErrorMessage(_result));
	  PyErr_ExceptionMatches(PoPy_OperationalError);
	  return NULL;
	}
      self->commit_on = PG_FALSE;
      PQclear(_result);
    }
  else 
    {
      PyErr_SetString(PoPy_OperationalError,"auto-commit mode or connection already closed");
      PyErr_ExceptionMatches(PoPy_OperationalError);
      return NULL;
    }
  Py_INCREF(Py_None);
  return Py_None;
}

static char PoPy_close__doc__[]="Connection Object method close";

/* Closing a connection without deleting the _ConnectionObject */
/* TODO: Does CO must be deleted ? ... No answer at now ! */
PyObject*
PoPy__ConnectionObject_close(self)
     PoPy__ConnectionObject *self;
{
  if(self->open) {
    PoPy__ConnectionObject_rollback(self);
    PQfinish(self->conn);
    self->open = PG_FALSE;
  }
  
  Py_INCREF(Py_None);
  return Py_None;
}

PyObject*
PoPy__ConnectionObject_cursor(self)
     PoPy__ConnectionObject *self;
{
  PyObject *cursor;

  cursor = (PyObject *) PoPy__CursorObject_NEW(self->conn,self->auto_commit);
  if (cursor == NULL) return NULL;

  return cursor;
}


static PyMethodDef PoPy__ConnectionObject_methods[] = {
  {"close",(PyCFunction)PoPy__ConnectionObject_close,0,PoPy_close__doc__},
  {"autoCommit",(PyCFunction)PoPy__ConnectionObject_autoCommit,METH_VARARGS},
  {"commit",(PyCFunction)PoPy__ConnectionObject_commit,0,PoPy_commit__doc__},
  {"cursor",(PyCFunction)PoPy__ConnectionObject_cursor,0,0},
  {"rollback",(PyCFunction)PoPy__ConnectionObject_rollback,0,0},
  {NULL, NULL}
};
      
static struct memberlist PoPy__ConnectionObject_memberlist[] = {
/*   {"cursor",T_OBJECT, offsetof(PoPy__ConnectionObject,cursor)}, */
  {"opened", T_INT, 0, RO},
  {"closed", T_INT, 0, RO},
  {"auto_commit_mode", T_STRING,0,RO},
  {NULL} /* Sentinel */
};

/* CO set attributes interface */
/* may disappear soon ! */
static int
PoPy__ConnectionObject_setattr(c, name, v)
	PoPy__ConnectionObject *c;
	char *name;
	PyObject *v;
{
	if (v == NULL) {
		PyErr_SetString(PyExc_AttributeError,
				"can't delete connection attributes");
		return -1;
	}
	return PyMember_Set((char *)c, PoPy__ConnectionObject_memberlist, name, v);
}



/* CO get attributes interface */
/* Very important method !!! */
static PyObject *
PoPy__ConnectionObject_getattr(self, name)
     PoPy__ConnectionObject *self;
     char *name;
{
	PyObject *_result;
	_result = Py_FindMethod(PoPy__ConnectionObject_methods, (PyObject *)self, name);
	if (_result != NULL)
		return _result;
	PyErr_Clear();
	if (strcmp(name, "opened") == 0)
	  return PyInt_FromLong((long)(self->open));
	if (strcmp(name, "closed") == 0)
	  return PyInt_FromLong((long)!(self->open));
	if (strcmp(name,"auto_commit_mode") == 0)
	  return PyString_FromString(self->auto_commit?"ON":"OFF");
	return PyMember_Get((char *)self, PoPy__ConnectionObject_memberlist, name);
}

/* To create new exceptions */
static PyObject *
PoPy_NewException(PyObject *dict,
		    char *name,
		    PyObject *base)
{
	PyObject *exc;
	char longname[256];

	sprintf(longname, "PoPy.%s", name);
	if ((exc = PyErr_NewException(longname, base, NULL)) == NULL)
		return NULL;
	if (PyDict_SetItemString(dict, name, exc)) return NULL;
	return exc;
}


/****************************************/
/*   DATE / TIME  METHODS DECLARATION   */
/****************************************/

PyObject * 
Date(self,args)
     PyObject *self;
     PyObject *args;
{
  PyObject * _date;
  int _year,_month,_day;
  
  if(!PyArg_ParseTuple(args,"iii",&_year,&_month,&_day))
    {
      PyErr_SetString(PoPy_ProgrammingError,"Bad Arguments");
      PyErr_ExceptionMatches(PoPy_ProgrammingError);
      
      return NULL;
    }
  
  if(mxDateTime_ImportModuleAndAPI() == -1)
    {
      PyErr_SetString(PoPy_InterfaceError,"Cannot load module DateTime");
      PyErr_ExceptionMatches(PoPy_InterfaceError);
      return NULL;
    }

  _date = mxDateTime.DateTime_FromDateAndTime(_year,_month,_day,0,0,0.0);
  
  if(!_date)
    {
      PyErr_SetString(PoPy_InterfaceError,"Creation of date failed");
      PyErr_ExceptionMatches(PoPy_InterfaceError);
      return NULL;
    }
  
  return _date;
} 

PyObject * 
Time(self,args)
     PyObject *self;
     PyObject *args;
{
  PyObject * _time;
  int _hours,_minutes;
  double _seconds;
  
  if(!PyArg_ParseTuple(args,"iid",&_hours,&_minutes,&_seconds))
    {
      PyErr_SetString(PoPy_ProgrammingError,"Bad Arguments");
      PyErr_ExceptionMatches(PoPy_ProgrammingError);
      
      return NULL;
    }
  
  if(mxDateTime_ImportModuleAndAPI() == -1)
    {
      PyErr_SetString(PoPy_InterfaceError,"Cannot load module DateTime");
      PyErr_ExceptionMatches(PoPy_InterfaceError);
      return NULL;
    }
  _time = mxDateTime.DateTimeDelta_FromTime(_hours,_minutes,_seconds);
  
  if(!_time)
    {
      PyErr_SetString(PoPy_InterfaceError,"Creation of time failed");
      PyErr_ExceptionMatches(PoPy_InterfaceError);
      return NULL;
    }
  
  return _time;
}

PyObject * 
Timestamp(self,args)
     PyObject *self;
     PyObject *args;
{
  PyObject * _timestamp;
  int _year,_month,_day,_hours,_minutes;
  double _seconds;
  
  if(!PyArg_ParseTuple(args,"iiiiid",&_year,&_month,&_day,&_hours,&_minutes,&_seconds))
    {
      PyErr_SetString(PoPy_ProgrammingError,"Bad Arguments");
      PyErr_ExceptionMatches(PoPy_ProgrammingError);
      return NULL;
    }
  
  if(mxDateTime_ImportModuleAndAPI() == -1)
    {
      PyErr_SetString(PoPy_ProgrammingError,"Cannot load module DateTime");
      PyErr_ExceptionMatches(PoPy_ProgrammingError);
      return NULL;
    }
  _timestamp = mxDateTime.DateTime_FromDateAndTime(_year,_month,_day,_hours,_minutes,_seconds);
  
  if(!_timestamp)
    {
      PyErr_SetString(PoPy_InterfaceError,"Creation of timestamp failed");
      PyErr_ExceptionMatches(PoPy_InterfaceError);
      return NULL;
    }
  
  return _timestamp;
}

PyObject * 
DateFromTicks(self,args)
     PyObject *self;
     PyObject *args;
{
  PyObject * _date;
  double _ticks;
  struct tm * _tm;
  time_t _tticks; 

  if(!PyArg_ParseTuple(args,"d",&_ticks))
    {
      PyErr_SetString(PoPy_ProgrammingError,"Bad Arguments");
      PyErr_ExceptionMatches(PoPy_ProgrammingError);
      return NULL;
    }

  if(mxDateTime_ImportModuleAndAPI() == -1)
   {
      PyErr_SetString(PoPy_ProgrammingError,"Cannot load module DateTime");
      PyErr_ExceptionMatches(PoPy_ProgrammingError);
      return NULL;
   }

  _tticks = (time_t) _ticks;
  _tm = localtime(&_tticks);

  _tm->tm_sec = 0;
  _tm->tm_min = 0;
  _tm->tm_hour = 0;

  _date = mxDateTime.DateTime_FromTmStruct(_tm);
  
  if(!_date)
    {
      PyErr_SetString(PoPy_InterfaceError,"Creation of date failed");
      PyErr_ExceptionMatches(PoPy_InterfaceError);
      return NULL;
    }
  return _date;
}

PyObject * 
TimeFromTicks(self,args)
     PyObject *self;
     PyObject *args;
{
  PyObject * _date;
  double _ticks;
  struct tm * _tm;
  time_t _tticks; 

  if(!PyArg_ParseTuple(args,"d",&_ticks))
    {
      PyErr_SetString(PoPy_ProgrammingError,"Bad Arguments");
      PyErr_ExceptionMatches(PoPy_ProgrammingError);
      return NULL;
    }
  
  if(mxDateTime_ImportModuleAndAPI() == -1)
    {
     PyErr_SetString(PoPy_ProgrammingError,"Cannot load module DateTime");
     PyErr_ExceptionMatches(PoPy_ProgrammingError);
     return NULL;
   }
  
  _tticks = (time_t) _ticks;
  _tm = localtime(&_tticks);

  _date = mxDateTime.DateTimeDelta_FromTime(_tm->tm_hour,_tm->tm_min,_tm->tm_sec);
  
  if(!_date)
    {
      PyErr_SetString(PoPy_InterfaceError,"Creation of date failed");
      PyErr_ExceptionMatches(PoPy_InterfaceError);
      return NULL;
    }
  return _date;
}

PyObject * 
TimestampFromTicks(self,args)
     PyObject *self;
     PyObject *args;
{
  PyObject * _date;
  double _ticks;
  
  if(!PyArg_ParseTuple(args,"d",&_ticks))
    {
      PyErr_SetString(PoPy_ProgrammingError,"Bad Arguments");
      PyErr_ExceptionMatches(PoPy_ProgrammingError);
      return NULL;
    }
  
  if(mxDateTime_ImportModuleAndAPI() == -1)
    {
      PyErr_SetString(PoPy_ProgrammingError,"Cannot load module DateTime");
      PyErr_ExceptionMatches(PoPy_ProgrammingError);
      return NULL;
    }
  
  _date = mxDateTime.DateTime_FromTicks(_ticks);
  
  if(!_date)
    {
      PyErr_SetString(PoPy_InterfaceError,"Creation of date failed");
      PyErr_ExceptionMatches(PoPy_InterfaceError);
      return NULL;
    }
  return _date;
}





/* ***************************************************** */
/*            -- MODULE FUNCTIONS DEFINITION --          */
/* ***************************************************** */

#ifdef __DEBUG__
static PyObject*
PoPy_test(self,args)
     PyObject *self;
     PyObject *args;
{
  PyObject *pdict,*res=NULL,*s,*d;
  
  PyArg_ParseTuple(args,"SO",&s,&d);

  pdict = PyDict_New();
  PyDict_SetItemString(pdict, "__builtins__", PyEval_GetBuiltins()); 
  PyDict_SetItemString(pdict, "format", s); 
  PyDict_SetItemString(pdict, "dict", d); 
  
  // extend the searchpath
  PyRun_String("import sys", Py_single_input, pdict, pdict); 
  PyRun_String("print dict", Py_single_input, pdict, pdict); 
  PyRun_String("format = format % dict", Py_single_input, pdict, pdict); 
  
  printf("result <%s>\n",PyString_AsString(PyDict_GetItemString(pdict,"format")));
/*   if(res == NULL) printf("Ther is an error !\n"); */
  Py_INCREF(Py_None);
  return Py_None;
}
#endif

/* Implementation of the main function of the module */
/* It allows to connect to a PostgreSQL database */
static PyObject*
PoPy_connect(self,args)
     PyObject *self;
     PyObject *args;
{
  return (PyObject *)PoPy__ConnectionObject_NEW(args);
}


static char PoPy___doc__[] =
"Adaptation of the PostgreSQL libpq library, and I try to
respect the Python DB-API specifications version 2.0, I hope ;).";

static PyMethodDef PoPy_methods[] = {
  {"connect",PoPy_connect, METH_VARARGS},
#ifdef __DEBUG__
  {"myTest",(PyCFunction)PoPy_test,METH_VARARGS},
#endif
  {"Binary",(PyCFunction)PoPy__BinaryObject_NEW,METH_VARARGS},
  {"Date",Date, METH_VARARGS},
  {"Time",Time, METH_VARARGS},
  {"Timestamp",Timestamp, METH_VARARGS},
  {"DateFromTicks",DateFromTicks, METH_VARARGS},
  {"TimeFromTicks",TimeFromTicks, METH_VARARGS},
  {"TimestampFromTicks",TimestampFromTicks, METH_VARARGS},
  {NULL, NULL}
};


/***********************************/		   
/*    OBJECTS TYPES DEFINITIONS    */
/***********************************/		   

PyTypeObject PoPy__ConnectionObject_Type = {
  PyObject_HEAD_INIT(&PyType_Type)
  0,
  "_ConnectionObject",
  sizeof(PoPy__ConnectionObject),
  0,
  (destructor)PoPy__ConnectionObject_dealloc, /* tp_dealloc */
  0, /*tp_print*/
  (getattrfunc)PoPy__ConnectionObject_getattr, /* tp_getattr */ 
  (setattrfunc)PoPy__ConnectionObject_setattr, /* tp_setattr */
  0, /*tp_compare*/
  (reprfunc)PoPy__ConnectionObject_repr, /* tp_repr */
 /*  , */
/*   PoPy__ConnectionObject_Type__doc__, */
};                         


PyTypeObject PoPy__CursorObject_Type = {
  PyObject_HEAD_INIT(&PyType_Type)
  0,
  "_CursorObject",
  sizeof(PoPy__CursorObject),
  0,
  (destructor)PoPy__CursorObject_dealloc, /* tp_dealloc */
  0, /*tp_print*/
  (getattrfunc)PoPy__CursorObject_getattr, /* tp_getattr */ 
  (setattrfunc)PoPy__CursorObject_setattr, /* tp_setattr */
  0, /*tp_compare*/
  (reprfunc)PoPy__CursorObject_repr, /* tp_repr */
};  


PyTypeObject PoPy__BinaryObject_Type = {
        PyObject_HEAD_INIT(&PyType_Type)
        0,
        "binary",
        sizeof(PoPy__BinaryObject),
        0,
        (destructor)PoPy__BinaryObject_dealloc, /*tp_dealloc*/
        0,              /*tp_print*/
        (getattrfunc)PoPy__BinaryObject_getattr,              /*tp_getattr*/
        0,              /*tp_setattr*/
        0, /*tp_compare*/
        (reprfunc)PoPy__BinaryObject_repr, /*tp_repr*/
        0,              /*tp_as_number*/
        0,    /*tp_as_sequence*/
        0,              /*tp_as_mapping*/
        0,  /*tp_hash*/
        0,              /*tp_call*/
        (reprfunc)PoPy__BinaryObject_repr,           /*tp_str*/
        0,              /*tp_getattro*/
        0,              /*tp_setattro*/
        0,      /*tp_as_buffer*/
        Py_TPFLAGS_DEFAULT,     /*tp_flags*/
        0,              /*tp_doc*/
};



/***************************************/
/*    PoPy MODULE INITIALIZATION    */
/***************************************/

void 
initPoPy()
{
  PyObject *dict, *module;
  
  module = Py_InitModule3("PoPy",PoPy_methods,PoPy___doc__);
  dict = PyModule_GetDict(module);

  PyDict_SetItemString(dict,"__version__",PyString_FromString("1.2"));
  PyDict_SetItemString(dict,"apilevel",PyString_FromString("2.0"));
  PyDict_SetItemString(dict,"threadsafety",PyInt_FromLong(1L));
  PyDict_SetItemString(dict,"paramstyle",PyString_FromString("pyformat"));
  PyDict_SetItemString(dict,"INV_READ",PyLong_FromLong(INV_READ));
  PyDict_SetItemString(dict,"INV_WRITE",PyLong_FromLong(INV_WRITE));
/*   PyDict_SetItemString(dict,"INV_ARCHIVE",PyLong_FromLong(INV_ARCHIVE)); */


  if (!(PoPy_Error = PoPy_NewException(dict, "Error", PyExc_StandardError)))
    goto error;
  if(!(PoPy_Warning = PoPy_NewException(dict, "Warning", PyExc_StandardError))) goto error;
if(!( PoPy_InterfaceError = PoPy_NewException(dict, "InterfaceError", PoPy_Error)))goto error;
  if(!(PoPy_DatabaseError = PoPy_NewException(dict, "DatabaseError", PoPy_Error)))goto error;
  if(!(PoPy_DataError = PoPy_NewException(dict, "DataError", PoPy_DatabaseError)))goto error;
  if(!(PoPy_OperationalError = PoPy_NewException(dict, "OperationalError", PoPy_DatabaseError)))goto error;
  if(!(PoPy_IntegrityError = PoPy_NewException(dict, "IntegrityError", PoPy_DatabaseError)))goto error;
  if(!(PoPy_InternalError = PoPy_NewException(dict, "InternalError", PoPy_DatabaseError)))goto error;
  if(!(PoPy_ProgrammingError = PoPy_NewException(dict, "ProgrammingError", PoPy_DatabaseError)))goto error;
  if(!(PoPy_NotSupportedError = PoPy_NewException(dict, "NotSupportedError", PoPy_DatabaseError)))goto error;
 error:
  if (PyErr_Occurred())
    PyErr_SetString(PyExc_ImportError,"PoPy: init failed");
  return;
}
  
